import sys
from time import sleep, time
import serial
import argparse, platform, re
if platform.system() == 'Linux':
    from serial.tools import list_ports_linux
    _list_ports = list_ports_linux
elif platform.system() == 'Windows':
    from serial.tools import list_ports_windows
    _list_ports = list_ports_windows
elif platform.system() == 'Darwin':
    from serial.tools import list_ports_osx
    _list_ports = list_ports_osx

MODULE_NAME = 'HL7588' #'HL7588' 'HL7648'


def detect_device(port: str = None):
    if port:
        return serial.Serial(port)
    _found = False
    print("Searching for {} modem.".format(MODULE_NAME))
    _f_tries = 0
    while True:
        _f_tries += 1
        for device in _list_ports.comports():
            try:
                port = serial.Serial(
                    port=device.device,
                    baudrate=115200,
                    parity=serial.PARITY_NONE,
                    stopbits=serial.STOPBITS_ONE,
                    bytesize=serial.EIGHTBITS,
                    rtscts=True,
                    dsrdtr=True
                )
                port.reset_input_buffer()
                port.reset_output_buffer()
                port.write('ATI\r\n'.encode())
                start = time()
                while not port.in_waiting:
                    sleep(.03)
                    if (time()-start) > 5:
                        port.close()
                        continue
                output = wait_for_output(port=port, str_wait='OK')
                if MODULE_NAME in output:
                    _found = True
                    print("Found {}.".format(MODULE_NAME))
                    return port
            except:
                continue
        sleep(2)
        if _f_tries > 10:
            print("Failed to find {} modem.".format(MODULE_NAME))
            sys.exit(1)
        if not _found:
            print("Failed to find {} modem.".format(MODULE_NAME))
            return False


def wait_for_output(port, str_wait='OK', timeout=5):
    sleep(.5)
    start_time = time()
    while True:
        data = str(port.read_all())
        if str_wait in data.strip():
            port.reset_output_buffer()
            port.reset_input_buffer()
            return data.strip()
        elif 'ERROR' in data.strip():
            print(data.strip())
            return False
        if time() - start_time >= timeout:
            return False


def check_port(port):
    if not port.is_open:
        print("COM port is not open. Attempting to re-open.")
        port.open()


def check_reg(port: object):
    while True:
        port.write('AT+CGATT?\r\n'.encode())
        cgatt_out = wait_for_output(port=port, str_wait='+CGATT: ', timeout=args.reg_timeout)
        if cgatt_out == False:
            print("Timeout or error on command AT+CGATT.")
            return False
        if '+CGATT: 1' not in cgatt_out:
            reg = False
            break
        else:
            reg = True
            break
    return reg


def register_on_nw(port):
    port.write('AT+COPS=1,2,"00101"\r\n'.encode())
    if not wait_for_output(port=port):
        print('Timeout or error while attempting to register')


def deregister(port):
    print("De-registering from network.")
    check_port(port)
    port.write('AT+COPS=2\r\n'.encode())
    sleep(15)
    if not wait_for_output(port=port):
        print('Timeout or error on command AT+COPS=2.')
        return False
    return True


def main_loop(port, args):
    if not check_reg(port):
        register_on_nw(port)
    else:
        deregister(port)
        register_on_nw(port)
    check_port(port)
    print("Waiting for network registration...")
    _reg = check_reg(port)
    if _reg:
        print("Registered.")
        if args.apn is not None:
            port.write('AT+CGDCONT=3\r\n'.encode())
            sleep(1)
            port.write('AT+CGDCONT=3,"IPV4V6","{}"\r\n'.format(args.apn).encode())
            if not wait_for_output(port=port):
                print('Timeout or error on command AT+CGDCONT.')
                return False
    else:
        print("Failed to register.")
        return False
    print("Sleeping for {} seconds.".format(args.idle))
    sleep(args.idle)
    if args.data:
        if not setup_udp_conn(port, args):
            clear_udp_conns(port)
            return False
        print("Sending UDP data.")
        port.write('AT+KUDPSND= 1,"8.8.8.8",32,64\r\n'.encode())
        sleep(.1)
        if not wait_for_output(port=port, str_wait='CONNECT'):
            print("Failed to send UDP data.")
            clear_udp_conns(port)
            return
        port.write(bytes(64))
        port.write('--EOF--Pattern--'.encode())
        if not wait_for_output(port=port):
            print("Failed to send UDP data.")
        else:
            print("Data sent.")
        sleep(1)
        clear_udp_conns(port)
    deregister(port)


def setup_udp_conn(port, args):
    port.write('AT+KCNXCFG=1,"GPRS","{}"\r\n'.format(args.apn).encode())
    sleep(1)
    if not wait_for_output(port=port):
        print("Failed to set up udp connection.")
        return False
    port.write('AT+KUDPCFG=1,0\r\n'.encode())
    sleep(1)
    if not wait_for_output(port=port, str_wait='+KUDPCFG:', timeout=10):
        print("Error setting up UDP connection.")
        return False
    return True


def clear_udp_conns(port):
    port.write('AT+KUDPCFG?\r\n'.encode())
    while not port.in_waiting:
        sleep(.25)
        pass
    udp_conns = port.read_all()
    m = re.findall('\+KUDPCFG:\s(\d).*', udp_conns.decode('utf-8'))
    for id in m:
        port.write('AT+KUDPCLOSE={},0\r\n'.format(id).encode())
    udp_conns = None
    port.flushInput()
    port.flushOutput()


if __name__ == '__main__':
    print("Cell Detect test script.")
    parser = argparse.ArgumentParser(description='Simple registration and text message script for {} modem.'.format(MODULE_NAME),
                                     formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument('reg_timeout', metavar='t', type=int, help='Seconds to wait for successful network registration.')
    parser.add_argument('idle', metavar='s', type=int,
                        help='Seconds to sleep after registration, before additional actions are executed such as text or data transmission.')
    parser.add_argument('loop_times', metavar='l', type=int, help='Times to repeat main loop. 0 = loop forever.')
    parser.add_argument('--apn', type=str, help='APN for cellular carrier. Defaults to "agilent".')
    parser.add_argument('--data', action='store_true', help='Send UDP data after the idle period.')
    # parser.add_argument('--module', type=str, help='Module name')

    args = parser.parse_args()
    if args.apn is None:
        print("Setting APN to default, 'agilent'.")
        args.apn = 'agilent'
    dev = detect_device('COM99')
    if not dev:
        print("Failed to connect to {} modem.".format(MODULE_NAME))
    print("Setting up module...")
    dev.write('AT+CSCS="IRA"\r\n'.encode())
    sleep(1)
    dev.read_all()
    loops = args.loop_times
    if dev:
        if args.data:
            if not args.apn:
                print("In order to send UDP data, you must provide the correct APN.")
            clear_udp_conns(dev)
        if loops == 0:
            print("Looping forever.")
            while True:
                try:
                    main_loop(dev, args)
                except Exception as ee:
                    del dev
                    dev = detect_device('COM99')
        else:
            print("Looping {} times.".format(loops))
            for x in range(0, loops):
                try:
                    main_loop(dev, args)
                except Exception as ee:
                    del dev
                    dev = detect_device('COM99')
            print("Done.")
        dev.write('AT+CFUN=1,1\r\n'.encode())

